" vim: ts=8 sw=4 sts=4 et foldenable foldmethod=marker foldcolumn=1

scriptencoding utf-8

" Search Constants {{{1

let s:start_chars = '_A-Za-z\u00a8\u00aa\u00ad\u00af\u00b2-\u00b5\u00b7-\u00ba\u00bc-\u00be\u00c0-\u00d6\u00d8-\u00f6\u00f8-\u02ff\u0370-\u167f\u1681-\u180d\u180f-\u1dbf\u1e00-\u1fff\u200b-\u200d\u202a-\u202e\u203f-\u2040\u2054\u2060-\u20cf\u2100-\u218f\u2460-\u24ff\u2776-\u2793\u2c00-\u2dff\u2e80-\u2fff\u3004-\u3007\u3021-\u302f\u3031-\ud7ff\uf900-\ufd3d\ufd40-\ufdcf\ufdf0-\ufe1f\ufe30-\ufe44\ufe47-\ufffd'

let s:search_chars = s:start_chars . '\u0300-\u036f\u1dc0-\u1dff\u20d0-\u20ff\ufe20-\ufe2f0-9'

let s:cpp_identifier = '[' . s:start_chars . '][' . s:search_chars . ']*'
"let s:cpp_keywords = '\<\(alignas\|alignof\|and\|and_eq\|asm\|auto\|bitand\|bitor\|bool\|break\|case\|catch\|char\|char16_t\|char32_t\|class\|compl\|const\|constexpr\|const_cast\|continue\|decltype\|default\|delete\|do\|double\|dynamic_cast\|else\|enum\|explicit\|export\|extern\|false\|float\|for\|friend\|goto\|if\|int\|long\|mutable\|namespace\|new\|noexcept\|not\|not_eq\|nullptr\|operator\|or\|or_eq\|register\|reinterpret_cast\|return\|short\|signed\|sizeof\|static\|static_assert\|static_cast\|struct\|switch\|template\|this\|thread_local\|throw\|true\|try\|typedef\|typeid\|typename\|union\|unsigned\|using\|virtual\|void\|volatile\|wchar_t\|while\|xor\|xor_eq\)\>'

let s:modifiers = '\(public\_s*\|private\_s*\|protected\_s*\|static\_s*\|volatile\_s*\|inline\_s*\)\='
let s:access_query = '\(public\_s*\|private\_s*\|protected\_s*\)\='

let s:cpp_keywords = '\<\(alignas\|alignof\|and\|and_eq\|asm\|auto\|bitand\|bitor\|break\|case\|catch\|compl\|constexpr\|const_cast\|continue\|decltype\|default\|delete\|do\|dynamic_cast\|else\|explicit\|export\|extern\|false\|for\|friend\|goto\|if\|new\|noexcept\|not\|not_eq\|nullptr\|operator\|or\|or_eq\|register\|reinterpret_cast\|return\|sizeof\|static_assert\|static_cast\|switch\|template\|this\|thread_local\|throw\|true\|try\|typedef\|typeid\|typename\|union\|using\|volatile\|wchar_t\|while\|xor\|xor_eq\)\>'

let s:template_identifier = s:cpp_identifier . '\(\[\]\|<[:<>?,.[:space:]' . s:search_chars . ']*>\)\='
let s:namespace_identifier = '\(' . s:cpp_identifier . '[:' . s:search_chars . ']*::\)\='

let s:class = '\<\(namespace\|class\|struct\|enum\)\>'

let s:struct = '^\s*' . s:class . '\_s\+' . s:cpp_identifier . '\_[^{]*{'
let s:common = '^[^(]*' . s:namespace_identifier . s:template_identifier . '\_s\+' . s:namespace_identifier . s:cpp_identifier . '\_s*('

let s:class_query = '^\s*' . s:struct
let s:func_query = '^\s*' . s:common

let s:tag_query = '\(' . s:struct . '\|' . s:common . '\)'

" Local Functions {{{1
" Universal Functions {{{2
" General {{{3

function! s:isBefore(x,y)
    if a:x[0] < a:y[0] || (a:x[0] == a:y[0] && a:x[1] < a:y[1])
        return 1
    endif
    return 0
endfunction

function! s:contains(range,line)
    if a:line >= a:range[0] && a:line <= a:range[1]
        return 1
    endif
    return 0
endfunction

function! s:isSmallerRange(x,y)
    if (a:x[1] - a:x[0]) < (a:y[1] - a:y[0])
        return 1
    endif
    return 0
endfunction

function! s:trim(string)
    return substitute(a:string,'\(^\s*\|\s*$\)','','g')
endfunction

function! s:merge(a,b)
    let a:i = 0
    let a:j = 0
    let a:c = []

    while a:i < len(a:a) || a:j < len(a:b)
        if a:j >= len(a:b)
            call add(a:c,a:a[a:i])
            let a:i += 1
        elseif a:i >= len(a:a)
            call add(a:c,a:b[a:j])
            let a:j += 1
        elseif a:j >= len(a:b) || a:a[a:i] < a:b[a:j]
            call add(a:c,a:a[a:i])
            let a:i += 1
        elseif a:i >= len(a:a) || a:b[a:j] < a:a[a:i]
            call add(a:c,a:b[a:j])
            let a:j += 1
        else
            call add(a:c,a:a[a:i])
            let a:i += 1
            let a:j += 1
        endif
    endwhile
    return a:c
endfunction

function! s:compare(x,y)
    if a:x[0] < a:y[0]
        return -1
    elseif a:x[0] > a:y[0]
        return 1
    else
        if a:x[1] > a:y[1]
            return -1
        elseif a:x[1] < a:y[1]
            return 1
        else
            return 0
        endif
    endif
endfunction

" File Navigation {{{3

function! s:isAlone(...)
    let a:file = a:0 > 0 ? a:1 : expand('%:p')
    let a:count = 0
    for buf in getbufinfo()
        if buf['name'] == a:file
            if len(buf['windows']) > 1
                return 0
            endif
            return 1
        endif
    endfor
    return 1
endfunction

function! s:safeClose(...)
    let a:prev = 0
    let a:file = a:0 > 0 ? a:1 : expand('%:p')
    if getbufinfo(a:file)[0]['loaded'] == 1 && winnr("$") == 1 && tabpagenr("$") > 1 && tabpagenr() > 1 && tabpagenr() < tabpagenr("$")
        let a:prev = 1
    endif

    if index(s:open_bufs,a:file) < 0 && s:isAlone(a:file) == 1
        execute 'bwipeout ' a:file
    elseif a:file == expand('%:p')
        q
    endif

    if a:prev == 1
        tabprev
    endif
endfunction

function! s:findTags(temp_file,search_string,append)
    let a:fout = a:append == 'yes' ? '>>' : '>'
    call system('find ' . getcwd() . g:factorus_ignore_string . '-exec grep -l "' . a:search_string . '" {} + ' . a:fout . ' ' . a:temp_file . ' 2> /dev/null')
endfunction

function! s:narrowTags(temp_file,search_string)
    let a:n_temp_file = a:temp_file . '.narrow'
    call system('cat ' . a:temp_file . ' | xargs grep -l "' . a:search_string . '" {} + > ' . a:n_temp_file)
    call system('mv ' . a:n_temp_file . ' ' . a:temp_file)
endfunction

function! s:updateQuickFix(temp_file,search_string)
    let a:res = split(system('cat ' . a:temp_file . ' | xargs grep -n "' . a:search_string . '"'),'\n')
    call map(a:res,{n,val -> split(val,':')})
    call map(a:res,{n,val -> {'filename' : val[0], 'lnum' : val[1], 'text' : s:trim(join(val[2:],':'))}})
    let g:factorus_qf += a:res
endfunction

function! s:setQuickFix(type)
    let a:title = a:type . ' : '
    if g:factorus_show_changes == 1
        let a:title .= 'ChangedFiles'
    elseif g:factorus_show_changes == 2
        let a:title .= 'UnchangedFiles'
    else
        let a:title .= 'AllFiles'
    endif

    call setqflist(s:qf)
    call setqflist(s:qf,'r',{'title' : a:title})
endfunction

function! s:getUnchanged(search)
    let a:qf = []

    let a:temp_file = '.FactorusUnchanged'
    call s:findTags(a:temp_file,a:search,'no')

    let a:count = 0
    for file in readfile(a:temp_file)
        let a:lines = split(system('grep -n "' . a:search . '" ' . file),'\n')  

        let a:count += len(a:lines)
        for line in a:lines
            let a:un = split(line,':')
            call add(a:qf,{'lnum' : a:un[0], 'filename' : file, 'text' : s:trim(join(a:un[1:],''))})
        endfor
    endfor

    call system('rm -rf ' . a:temp_file)
    return a:qf
endfunction

" Utilities {{{2

function! s:getClosingBracket(stack,...)
    let a:orig = [line('.'),col('.')]
    if a:0 > 0
        call cursor(a:1[0],a:1[1])
    endif
    if a:stack == 0
        call searchpair('{','','}','Wb')
    else
        call search('{','Wc')
    endif
    execute 'normal %'
    let a:res = [line('.'),col('.')]
    call cursor(a:orig[0],a:orig[1])
    return a:res
endfunction

function! s:isQuoted(pat,state)
    let a:temp = a:state
    let a:mat = match(a:temp,a:pat)
    let a:res = 1
    while a:mat >= 0 && a:res == 1
        let a:begin = strpart(a:temp,0,a:mat)
        let a:quotes = len(a:begin) - len(substitute(a:begin,'"','','g'))
        if a:quotes % 2 == 1
            let a:res = 1
        else
            let a:res = 0
        endif
        let a:temp = substitute(a:temp,a:pat,'','')
        let a:mat = match(a:temp,a:pat)
    endwhile
    return a:res
endfunction

function! s:isWrapped(pat,state)
    let a:match = match(a:state,a:pat)
    let a:temp = a:state
    let a:res = 1
    while a:match >= 0
        let a:begin = split(strpart(a:temp,0,a:match),'\zs')
        if count(a:begin,'>') >= count(a:begin,'<')
            let a:res = 0
            break
        endif
        let a:temp = substitute(a:temp,a:pat,'','')
        let a:match = match(a:temp,a:pat)
    endwhile
    return a:res
endfunction

function! s:isCommented()
    if match(getline('.'),'//') >= 0 && match(getline('.'),'//') < col('.')
        return 1
    endif
    if searchpairpos('[^/]\/\*','','\*\/','Wbn') != [0,0]
        return 1
    endif
    return 0
endfunction

function! s:getEndLine(start,search)
    let a:orig = [line('.'),col('.')]
    call cursor(a:start[0],a:start[1])
    let a:fin = searchpos(a:search,'Wen')
    call cursor(a:orig[0],a:orig[1])
    return a:fin
endfunction

function! s:getStatement(line)
    let a:i = a:line
    while match(getline(a:i),'\({\|;$\)') < 0
        let a:i += 1
    endwhile
    return join(getline(a:line,a:i))
endfunction

" Tag Navigation {{{2
" isValidTag {{{3
function! s:isValidTag(line)
    let a:first_char = strpart(substitute(getline(a:line),'\s*','','g'),0,1)   
    if a:first_char == '*' || a:first_char == '/'
        return 0
    endif

    let a:has_keyword = match(getline(a:line),s:cpp_keywords)
    if a:has_keyword >= 0 && s:isQuoted(s:cpp_keywords,getline(a:line)) == 0
        return 0
    endif

    return 1
endfunction

" getAdjacentTag {{{3
function! s:getAdjacentTag(dir)
    let [a:oline,a:ocol] = [line('.'),col('.')]
    call cursor(a:oline + 1,a:ocol)

    let a:func = searchpos(s:tag_query,'Wn' . a:dir)
    let a:is_valid = 0
    while a:func != [0,0]
        call cursor(a:func[0],a:func[1])

        let a:close = (s:isBefore(searchpos('{','Wn'),searchpos(';','Wn')) == 1) ? s:getClosingBracket(1,a:func) : s:getClosingBracket(0,a:func)
        let a:is_valid = s:isValidTag(a:func[0])

        if a:is_valid == 1
            let a:still_valid = 1
            let a:back = searchpos(s:tag_query,'Wnb')
            while a:back != [0,0]
                call cursor(a:back[0],a:back[1])
                if s:isValidTag(a:back[0])
                    let a:back_close = (s:isBefore(searchpos('{','Wn'),searchpos(';','Wn')) == 1) ? s:getClosingBracket(1,a:back) : s:getClosingBracket(0,a:back)
                    if match(getline(a:back[0]),s:class) < 0 && (s:isBefore(a:close,a:back_close) || a:close == a:back_close)
                        let a:still_valid = 0
                    endif
                    break
                endif
                let a:back = searchpos(s:tag_query,'Wnb')
            endwhile
            if a:still_valid == 1
                break
            endif
        endif
        let a:is_valid = 0

        call cursor(a:func[0],a:func[1])
        let a:func = searchpos(s:tag_query,'Wn' . a:dir)
    endwhile
    call cursor(a:oline,a:ocol)

    if a:is_valid == 1
        return a:func[0]
    endif
    return 0
endfunction

" getNextTag {{{3
function! s:getNextTag()
    return [s:getAdjacentTag(''),1]
endfunction

" getClassTag {{{3
function! s:getClassTag()
    let [a:line,a:col] = [line('.'),col('.')]
    call cursor(1,1)
    let a:class_tag = search(s:tag_query,'n')
    let a:tag_end = search(s:tag_query,'ne')
    call cursor(a:line,a:col)
    return [a:class_tag,a:tag_end]
endfunction

" gotoTag {{{3
function! cpp#factorus#gotoTag(head)
    let a:tag = a:head == 1 ? s:getClassTag()[0] : s:getAdjacentTag('b') 
    if a:tag != 0
        call cursor(a:tag,1)
    elseif a:head == 1
        call cursor(1,1)
    else
        echo 'No tag found'
    endif
endfunction

" Class Hierarchy {{{2

" Declarations {{{2

" References {{{2

" File-Updating {{{2

" Renaming {{{2
" renameArg {{{3

" renameClass {{{3

" renameField {{{3

" renameMethod {{{3

" Extraction {{{2

" Method-Building {{{2

" Rollback {{{2

" Global Functions {{{1
" addParam {{{2

" renameSomething {{{2
function! cpp#factorus#enameSomething(new_name,type,...)
    let a:orig = [line('.'),col('.')]
    let s:open_bufs = []
    let s:qf = []

    let a:prev_dir = getcwd()
    let a:buf_nrs = []
    for buf in getbufinfo()
        call add(s:open_bufs,buf['name'])
        call add(a:buf_nrs,buf['bufnr'])
    endfor
    let a:curr_buf = a:buf_nrs[index(s:open_bufs,expand('%:p'))]
    let a:buf_setting = &switchbuf

    execute 'silent cd ' . expand('%:p:h')
    let a:project_dir = g:factorus_project_dir == '' ? system('git rev-parse --show-toplevel') : g:factorus_project_dir
    execute 'silent cd ' a:project_dir

    let a:res = ''
    try
        if factorus#isRollback(a:000)
            let a:res = s:rollbackRename()
            let g:factorus_qf = []
        else
            let g:factorus_qf = []
            let Rename = function('s:rename' . a:type)
            let a:res = Rename(a:new_name)

            if g:factorus_show_changes > 0
                let s:qf = copy(g:factorus_qf)

                let a:ch = len(g:factorus_qf)
                let a:ch_i = a:ch == 1 ? ' instance ' : ' instances '
                let a:un = s:getUnchanged('\<' . a:res . '\>')
                let a:un_l = len(a:un)
                let a:un_i = a:un_l == 1 ? ' instance ' : ' instances '

                let a:first_line = a:ch . a:ch_i . 'modified'
                let a:first_line .= a:type == 'Arg' ? '.' : ', ' . a:un_l . a:un_i . 'left unmodified.'

                if g:factorus_show_changes > 1 && a:type != 'Arg'
                    let a:un = [{'pattern' : 'Unmodified'}] + a:un
                    if g:factorus_show_changes == 2
                        let s:qf = []
                    endif
                    let s:qf += a:un
                endif

                if g:factorus_show_changes % 2 == 1
                    let s:qf = [{'pattern' : 'Modified'}] + s:qf
                endif
                let s:qf = [{'text' : a:first_line,'pattern' : 'rename' . a:type}] + s:qf

                call s:setQuickFix(a:type)
            endif
        endif

        execute 'silent cd ' a:prev_dir
        if a:type != 'Class'
            let &switchbuf = 'useopen,usetab'
            execute 'silent sbuffer ' . a:curr_buf
            let &switchbuf = a:buf_setting
        endif
        call cursor(a:orig[0],a:orig[1])
 
        return a:res
    catch /.*/
        call system('rm -rf .Factorus*')
        execute 'silent cd ' a:prev_dir
        if a:type != 'Class'
            let &switchbuf = 'useopen,usetab'
            execute 'silent sbuffer ' . a:curr_buf
            let &switchbuf = a:buf_setting
        endif
        let a:err = match(v:exception,'^Factorus:') >= 0 ? v:exception : 'Factorus:' . v:exception
        throw a:err . ', at ' . v:throwpoint
    endtry
endfunction

" extractMethod {{{2

